import os
from pathlib import Path

from sklearn.exceptions import NotFittedError
from sklearn.utils import validation

from sail.utils.logging import configure_logger
from sail.utils.serialization import load_obj, save_obj

LOGGER = configure_logger(logger_name="SAILModel")


class SAILModel:
    def fit(self):
        raise NotImplementedError

    def partial_fit(self):
        raise NotImplementedError

    def predict(self):
        raise NotImplementedError

    def score(self):
        raise NotImplementedError

    def check_is_fitted(self, attributes=None, *args, **kwargs):
        """Indicate whether the SAIL model has been fit."""
        try:
            validation.check_is_fitted(
                estimator=self, attributes=attributes, *args, **kwargs
            )
            return True
        except NotFittedError:
            return False

    def save_model(self, model_folder, file_name="model"):
        """Saves the sail model to pickle format.

        Args:
            model_folder: String, PathLike, path to SavedModel.
        """
        Path(model_folder).mkdir(parents=True, exist_ok=True)
        save_obj(
            self,
            model_folder,
            file_name,
        )

        Path(os.path.join(model_folder, ".sailmodel")).touch()

        LOGGER.info("Model saved successfully.")

    @classmethod
    def load_model(cls, model_folder, file_name="model"):
        """Load the sail model from the mentioned folder location

        Args:
            model_folder: One of the following:
                - String or `pathlib.Path` object, path to the saved model

        Returns:
            BaseModel: Model
        """
        model = load_obj(model_folder, file_name)
        LOGGER.info("Model loaded successfully.")

        return model



import importlib
import os
from pathlib import Path
import torch
import numpy as np
from skorch.classifier import NeuralNetClassifier
from skorch.regressor import NeuralNetRegressor
from sklearn.exceptions import NotFittedError
from sklearn.utils import validation
from sail.models.base import SAILModel
from sail.utils.logging import configure_logger
from sail.utils.serialization import load_obj, save_obj
import copy

LOGGER = configure_logger(logger_name="PyTorch")


class TorchSerializationMixin:
    def save_model(self, model_folder):
        """
        Saves the module's parameters, history, and optimizer.

        Args:
            model_folder: String, PathLike, path to model.
        """
        Path(model_folder).mkdir(parents=True, exist_ok=True)
        self.save_params(  # type: ignore
            f_params=model_folder + "/weights.pkl",
            f_optimizer=model_folder + "/opt.pkl",
            f_history=model_folder + "/history.json",
        )

        save_obj(
            obj={
                "class": f"{self.__class__.__module__}.{self.__class__.__qualname__}",
            },
            location=model_folder,
            file_name="metadata",
            serialize_type="json",
        )

        save_obj(
            obj=self.get_params_for("module"),
            location=model_folder,
            file_name="params",
        )

        Path(os.path.join(model_folder, ".pytorch")).touch()

        LOGGER.info("Model saved successfully.")

    @classmethod
    def load_model(cls, model_folder):
        """
        Loads the the module's parameters, history, and optimizer.

        Args:
            model_folder: One of the following:
                - String or `pathlib.Path` object, path to the saved model
        """
        metadata = load_obj(
            location=model_folder,
            file_name="metadata",
            serialize_type="json",
        )

        params = load_obj(
            location=model_folder,
            file_name="params",
        )

        class_str = metadata["class"].split(".")[-1]
        module_path = ".".join(metadata["class"].split(".")[0:-1])
        model_class = getattr(importlib.import_module(module_path), class_str)
        model = model_class(**params)

        model.initialize()
        if os.path.exists(model_folder):
            model.load_params(  # type: ignore
                f_params=model_folder + "/weights.pkl",
                f_optimizer=model_folder + "/opt.pkl",
                f_history=model_folder + "/history.json",
            )

        return model


class TorchModelMixin:
    def fit(self, X, y=None, **fit_params):
        X, y = self.cast_X_y(X, y)
        return super().fit(X, y, **fit_params)

    def partial_fit(self, X, y=None, classes=None, **fit_params):
        X, y = self.cast_X_y(X, y)
        return super().partial_fit(X, y, classes, **fit_params)

    def predict(self, X):
        X = self.cast_X_y(X)
        return super().predict(X)

    def cast_X_y(self, X, y=None):
        if not isinstance(X, torch.Tensor):
            X = np.asarray(X).astype(np.float32)
        if y is not None:
            if not isinstance(y, torch.Tensor):
                y = np.asarray(y)
                if np.float64 == y.dtype:
                    y = y.astype(np.float32).reshape((-1,))
            return X, y
        else:
            return X

    def check_is_fitted(self, attributes=None, *args, **kwargs):
        """Indicate whether the torch model has been fit."""
        try:
            attributes = (
                attributes or [module + "_" for module in self._modules] or ["module_"]
            )

            validation.check_is_fitted(
                estimator=self, attributes=attributes, *args, **kwargs
            )
            return True
        except NotFittedError:
            return False

    def get_params(self, deep=True):
        """
        Get parameters for this estimator.

        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key)
            if deep and hasattr(value, "get_params") and not isinstance(value, type):
                deep_items = value.get_params().items()
                out.update((key + "__" + k, val) for k, val in deep_items)
            out[key] = value

        out.pop("module", None)
        for param in copy.deepcopy(out):
            if param.startswith("module__"):
                out.pop(param, None)

        return out


class SAILTorchRegressor(
    TorchModelMixin, NeuralNetRegressor, TorchSerializationMixin, SAILModel
):
    def __init__(self, *args, max_epochs=1, batch_size=-1, train_split=None, **kwargs):
        super(SAILTorchRegressor, self).__init__(
            *args,
            max_epochs=max_epochs,
            batch_size=batch_size,
            train_split=train_split,
            **kwargs,
        )


class SAILTorchClassifier(
    TorchModelMixin, NeuralNetClassifier, TorchSerializationMixin, SAILModel
):
    def __init__(
        self,
        *args,
        max_epochs=1,
        batch_size=-1,
        train_split=None,
        **kwargs,
    ):
        super(SAILTorchClassifier, self).__init__(
            *args,
            max_epochs=max_epochs,
            batch_size=batch_size,
            train_split=train_split,
            **kwargs,
        )



import torch
from torch import nn
import torch.nn.functional as F

from sklearn.datasets import make_classification
from sklearn.preprocessing import MinMaxScaler
import numpy as np


from typing import cast, Union, List

from sail.models.torch import SAILTorchClassifier

"""
Define conv1d with same padding
"""


class Conv1dSamePadding(nn.Conv1d):
    """Represents the "Same" padding functionality from Tensorflow.
    See: https://github.com/pytorch/pytorch/issues/3867
    Note that the padding argument in the initializer doesn't do anything now
    """

    def forward(self, input):
        return conv1d_same_padding(
            input, self.weight, self.bias, self.stride, self.dilation, self.groups
        )


def conv1d_same_padding(input, weight, bias, stride, dilation, groups):
    # stride and dilation are expected to be tuples.
    kernel, dilation, stride = weight.size(2), dilation[0], stride[0]
    l_out = l_in = input.size(2)
    padding = ((l_out - 1) * stride) - l_in + (dilation * (kernel - 1)) + 1
    if padding % 2 != 0:
        input = F.pad(input, [0, 1])

    return F.conv1d(
        input=input,
        weight=weight,
        bias=bias,
        stride=stride,
        padding=padding // 2,
        dilation=dilation,
        groups=groups,
    )


class ConvBlock(nn.Module):
    def __init__(
        self, in_channels: int, out_channels: int, kernel_size: int, stride: int
    ) -> None:
        super().__init__()

        self.layers = nn.Sequential(
            Conv1dSamePadding(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
            ),
            nn.BatchNorm1d(num_features=out_channels),
            nn.ReLU(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore
        return self.layers(x)


class _InceptionModel(nn.Module):
    """A PyTorch implementation of the InceptionTime model.
    From https://arxiv.org/abs/1909.04939
    Attributes
    ----------
    num_blocks:
        The number of inception blocks to use. One inception block consists
        of 3 convolutional layers, (optionally) a bottleneck and (optionally) a residual
        connector
    in_channels:
        The number of input channels (i.e. input.shape[-1])
    out_channels:
        The number of "hidden channels" to use. Can be a list (for each block) or an
        int, in which case the same value will be applied to each block
    bottleneck_channels:
        The number of channels to use for the bottleneck. Can be list or int. If 0, no
        bottleneck is applied
    kernel_sizes:
        The size of the kernels to use for each inception block. Within each block, each
        of the 3 convolutional layers will have kernel size
        `[kernel_size // (2 ** i) for i in range(3)]`
    num_pred_classes:
        The number of output classes
    """

    def __init__(
        self,
        num_blocks: int,
        in_channels: int,
        out_channels: Union[List[int], int],
        bottleneck_channels: Union[List[int], int],
        kernel_sizes: Union[List[int], int],
        use_residuals: Union[List[bool], bool, str] = "default",
        num_pred_classes: int = 1,
    ) -> None:
        super().__init__()

        # for easier saving and loading
        self.input_args = {
            "num_blocks": num_blocks,
            "in_channels": in_channels,
            "out_channels": out_channels,
            "bottleneck_channels": bottleneck_channels,
            "kernel_sizes": kernel_sizes,
            "use_residuals": use_residuals,
            "num_pred_classes": num_pred_classes,
        }

        channels = [in_channels] + cast(
            List[int], self._expand_to_blocks(out_channels, num_blocks)
        )
        bottleneck_channels = cast(
            List[int], self._expand_to_blocks(bottleneck_channels, num_blocks)
        )
        kernel_sizes = cast(List[int], self._expand_to_blocks(kernel_sizes, num_blocks))
        if use_residuals == "default":
            use_residuals = [True if i % 3 == 2 else False for i in range(num_blocks)]
        use_residuals = cast(
            List[bool],
            self._expand_to_blocks(
                cast(Union[bool, List[bool]], use_residuals), num_blocks
            ),
        )

        self.blocks = nn.Sequential(
            *[
                _InceptionBlock(
                    in_channels=channels[i],
                    out_channels=channels[i + 1],
                    residual=use_residuals[i],
                    bottleneck_channels=bottleneck_channels[i],
                    kernel_size=kernel_sizes[i],
                )
                for i in range(num_blocks)
            ]
        )

        # a global average pooling (i.e. mean of the time dimension) is why
        # in_features=channels[-1]
        self.linear = nn.Linear(in_features=channels[-1], out_features=num_pred_classes)

    @staticmethod
    def _expand_to_blocks(
        value: Union[int, bool, List[int], List[bool]], num_blocks: int
    ) -> Union[List[int], List[bool]]:
        if isinstance(value, list):
            assert len(value) == num_blocks, (
                f"Length of inputs lists must be the same as num blocks, "
                f"expected length {num_blocks}, got {len(value)}"
            )
        else:
            value = [value] * num_blocks
        return value

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore
        x = self.blocks(x).mean(dim=-1)  # the mean is the global average pooling
        return self.linear(x)


class _InceptionBlock(nn.Module):
    """An inception block consists of an (optional) bottleneck, followed
    by 3 conv1d layers. Optionally residual
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        residual: bool,
        stride: int = 1,
        bottleneck_channels: int = 32,
        kernel_size: int = 41,
    ) -> None:
        assert kernel_size > 3, "Kernel size must be strictly greater than 3"
        super().__init__()

        self.use_bottleneck = bottleneck_channels > 0
        if self.use_bottleneck:
            self.bottleneck = Conv1dSamePadding(
                in_channels, bottleneck_channels, kernel_size=1, bias=False
            )
        kernel_size_s = [kernel_size // (2**i) for i in range(3)]
        start_channels = bottleneck_channels if self.use_bottleneck else in_channels
        channels = [start_channels] + [out_channels] * 3
        self.conv_layers = nn.Sequential(
            *[
                Conv1dSamePadding(
                    in_channels=channels[i],
                    out_channels=channels[i + 1],
                    kernel_size=kernel_size_s[i],
                    stride=stride,
                    bias=False,
                )
                for i in range(len(kernel_size_s))
            ]
        )

        self.batchnorm = nn.BatchNorm1d(num_features=channels[-1])
        self.relu = nn.ReLU()

        self.use_residual = residual
        if residual:
            self.residual = nn.Sequential(
                *[
                    Conv1dSamePadding(
                        in_channels=in_channels,
                        out_channels=out_channels,
                        kernel_size=1,
                        stride=stride,
                        bias=False,
                    ),
                    nn.BatchNorm1d(out_channels),
                    nn.ReLU(),
                ]
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore
        org_x = x
        if self.use_bottleneck:
            x = self.bottleneck(x)
        x = self.conv_layers(x)

        if self.use_residual:
            x = x + self.residual(org_x)
        return x


"""
wrapper scorch class
"""


class InceptionTimeClassifier(SAILTorchClassifier):
    def __init__(
        self,
        num_blocks=2,
        in_channels=1,
        out_channels=2,
        bottleneck_channels=2,
        kernel_sizes=41,
        use_residuals=True,
        num_pred_classes=1,
        learning_rate=0.05,
        batch_size=1000,
        criterion=nn.BCEWithLogitsLoss,
        max_epochs=50,
        device="cpu",
    ):
        self.inception_model = _InceptionModel(
            num_blocks=2,
            in_channels=1,
            out_channels=2,
            bottleneck_channels=2,
            kernel_sizes=41,
            use_residuals=True,
            num_pred_classes=1,
        )

        super(InceptionTimeClassifier, self).__init__(
            module=self.inception_model,
            max_epochs=max_epochs,
            lr=learning_rate,
            batch_size=batch_size,
            criterion=criterion,
            # Shuffle training data on each epoch
            iterator_train__shuffle=False,
            device=device,
        )



import tensorflow as tf
from sail.models.keras import SAILKerasClassifier
from typing import List


class _Model:
    def __init__(
        self,
        num_hidden_nodes: List[int] = [100],
        hidden_layer_activation: List[str] = ["relu"],
        num_output_nodes: int = 1,
        outer_layer_activation: str = "sigmoid",
    ):
        self.num_hidden_nodes = num_hidden_nodes
        self.hidden_layer_activation = hidden_layer_activation
        self.num_output_nodes = num_output_nodes
        self.outer_layer_activation = outer_layer_activation

    def create_instance(self):
        model = tf.keras.models.Sequential()

        assert isinstance(self.num_hidden_nodes, list) and isinstance(
            self.hidden_layer_activation, list
        ), "num_hidden_nodes and hidden_layer_activation must be of the type List[str]."

        for num_hidden_nodes, hidden_layer_activation in zip(
            self.num_hidden_nodes, self.hidden_layer_activation
        ):
            model.add(
                tf.keras.layers.Dense(
                    num_hidden_nodes, activation=hidden_layer_activation
                )
            )

        model.add(
            tf.keras.layers.Dense(
                self.num_output_nodes, activation=self.outer_layer_activation
            )
        )
        return model


class KerasSequentialClassifier(SAILKerasClassifier):
    def __init__(
        self,
        num_hidden_nodes: int = 8,
        hidden_layer_activation: str = "relu",
        num_output_nodes: int = 1,
        outer_layer_activation: str = "sigmoid",
        loss="binary_crossentropy",
        optimizer=tf.keras.optimizers.Adam,
        metrics=["accuracy"],
        epochs=1,
        verbose=0,
        **kwargs,
    ) -> None:
        super(KerasSequentialClassifier, self).__init__(
            model=_Model(
                num_hidden_nodes,
                hidden_layer_activation,
                num_output_nodes,
                outer_layer_activation,
            ).create_instance(),
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            epochs=epochs,
            verbose=verbose,
        )
        self.num_hidden_nodes = num_hidden_nodes
        self.hidden_layer_activation = hidden_layer_activation
        self.num_output_nodes = num_output_nodes
        self.outer_layer_activation = outer_layer_activation



import os
from pathlib import Path
import tensorflow as tf
from scikeras.wrappers import KerasRegressor, KerasClassifier
import importlib
import inspect
import numpy as np
from scikeras.utils import loss_name
from sail.models.base import SAILModel
from sail.utils.logging import configure_logger
from sail.utils.serialization import load_obj, save_obj

LOGGER = configure_logger(logger_name="TF_Keras")


class KerasSerializationMixin:
    def save_model(self, model_folder):
        """
        Saves Model to Tensorflow SavedModel format.

        Args:
            model_folder: String, PathLike, path to SavedModel.
        """
        if self.model_ is not None:
            self.model_.save(model_folder, overwrite=True, save_format="tf")

            Path(os.path.join(model_folder, ".keras")).touch()

            save_obj(
                obj={
                    "class": f"{self.__class__.__module__}.{self.__class__.__qualname__}",
                },
                location=model_folder,
                file_name="metadata",
                serialize_type="json",
            )

            params = self.get_params()
            save_obj(
                obj=params,
                location=model_folder,
                file_name="params",
            )

            np.savez(
                os.path.join(model_folder, "init_data"),
                init_X=self.init_X,
                init_y=self.init_y,
            )

            LOGGER.info("Model saved successfully.")

    @classmethod
    def load_model(cls, model_folder):
        """
        Load Model from the mentioned folder location

        Args:
            model_folder: One of the following:
                - String or `pathlib.Path` object, path to the saved model
        """
        if os.path.exists(model_folder):
            keras_model = tf.keras.models.load_model(model_folder)

            metadata = load_obj(
                location=model_folder,
                file_name="metadata",
                serialize_type="json",
            )

            params = load_obj(
                location=model_folder,
                file_name="params",
            )

            class_str = metadata["class"].split(".")[-1]
            module_path = ".".join(metadata["class"].split(".")[0:-1])
            model_class = getattr(importlib.import_module(module_path), class_str)
            model_wrapper = model_class(**params)

            model_wrapper.model = keras_model
            init_data = np.load(os.path.join(model_folder, "init_data.npz"))
            model_wrapper.initialize(init_data["init_X"], init_data["init_y"])

            LOGGER.info("Model loaded successfully.")
            return model_wrapper


class KerasModelMixin:
    def _fit(
        self,
        X,
        y,
        sample_weight,
        warm_start: bool,
        epochs: int,
        initial_epoch: int,
        **kwargs,
    ):
        if not self.initialized_:
            self.init_X = X
            self.init_y = y

        if not ((self.warm_start or warm_start) and self.initialized_):
            X, y = self._initialize(X, y)
        else:
            X, y = self._validate_data(X, y)
        self._ensure_compiled_model()

        if sample_weight is not None:
            X, y, sample_weight = self._validate_sample_weight(X, y, sample_weight)

        y = self.target_encoder_.transform(y)
        X = self.feature_encoder_.transform(X)

        self._fit_keras_model(
            X,
            y,
            sample_weight=sample_weight,
            warm_start=warm_start,
            epochs=epochs,
            initial_epoch=initial_epoch,
            **kwargs,
        )

    def get_params(self, deep=True):
        """
        Get parameters for this estimator.

        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key)
            if deep and hasattr(value, "get_params") and not isinstance(value, type):
                deep_items = value.get_params().items()
                out.update((key + "__" + k, val) for k, val in deep_items)
            out[key] = value

        out.pop("model", None)
        out.pop("init_X", None)
        out.pop("init_y", None)
        out.pop("validation_split", None)
        return out


class SAILKerasRegressor(
    KerasModelMixin, KerasRegressor, KerasSerializationMixin, SAILModel
):
    def __init__(self, *args, **kwargs):
        super(SAILKerasRegressor, self).__init__(
            *args,
            **kwargs,
        )


class SAILKerasClassifier(
    KerasModelMixin, KerasClassifier, KerasSerializationMixin, SAILModel
):
    def __init__(self, *args, **kwargs):
        super(SAILKerasClassifier, self).__init__(
            *args,
            **kwargs,
        )



"""
Omni-Scale 1D-CNN

Code adapted from https://github.com/Wensi-Tang/OS-CNN
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import math
import numpy as np
import torch.nn as nn
import torch.optim as optim
from sail.models.torch import SAILTorchClassifier


def get_Prime_number_in_a_range(start, end):
    Prime_list = []
    for val in range(start, end + 1):
        prime_or_not = True
        for n in range(2, val):
            if (val % n) == 0:
                prime_or_not = False
                break
        if prime_or_not:
            Prime_list.append(val)
    return Prime_list


def get_out_channel_number(paramenter_layer, in_channel, prime_list):
    out_channel_expect = int(paramenter_layer / (in_channel * sum(prime_list)))
    return out_channel_expect


def generate_layer_parameter_list(
    start, end, paramenter_number_of_layer_list, in_channel=1
):
    prime_list = get_Prime_number_in_a_range(start, end)
    # if prime_list == []:
    #     # print('start = ',start, 'which is larger than end = ', end)
    input_in_channel = in_channel
    layer_parameter_list = []
    for paramenter_number_of_layer in paramenter_number_of_layer_list:
        out_channel = get_out_channel_number(
            paramenter_number_of_layer, in_channel, prime_list
        )

        tuples_in_layer = []
        for prime in prime_list:
            tuples_in_layer.append((in_channel, out_channel, prime))
        in_channel = len(prime_list) * out_channel

        layer_parameter_list.append(tuples_in_layer)

    tuples_in_layer_last = []
    first_out_channel = len(prime_list) * get_out_channel_number(
        paramenter_number_of_layer_list[0], input_in_channel, prime_list
    )
    tuples_in_layer_last.append((in_channel, first_out_channel, start))
    tuples_in_layer_last.append((in_channel, first_out_channel, start + 1))
    layer_parameter_list.append(tuples_in_layer_last)
    return layer_parameter_list


def calculate_mask_index(kernel_length_now, largest_kernel_lenght):
    right_zero_mast_length = math.ceil((largest_kernel_lenght - 1) / 2) - math.ceil(
        (kernel_length_now - 1) / 2
    )
    left_zero_mask_length = (
        largest_kernel_lenght - kernel_length_now - right_zero_mast_length
    )
    return left_zero_mask_length, left_zero_mask_length + kernel_length_now


def creat_mask(
    number_of_input_channel,
    number_of_output_channel,
    kernel_length_now,
    largest_kernel_lenght,
):
    ind_left, ind_right = calculate_mask_index(kernel_length_now, largest_kernel_lenght)
    mask = np.ones(
        (number_of_input_channel, number_of_output_channel, largest_kernel_lenght)
    )
    mask[:, :, 0:ind_left] = 0
    mask[:, :, ind_right:] = 0
    return mask


def creak_layer_mask(layer_parameter_list):
    largest_kernel_lenght = layer_parameter_list[-1][-1]
    mask_list = []
    init_weight_list = []
    bias_list = []
    for i in layer_parameter_list:
        conv = torch.nn.Conv1d(in_channels=i[0], out_channels=i[1], kernel_size=i[2])
        ind_l, ind_r = calculate_mask_index(i[2], largest_kernel_lenght)
        big_weight = np.zeros((i[1], i[0], largest_kernel_lenght))
        big_weight[:, :, ind_l:ind_r] = conv.weight.detach().numpy()

        bias_list.append(conv.bias.detach().numpy())
        init_weight_list.append(big_weight)

        mask = creat_mask(i[1], i[0], i[2], largest_kernel_lenght)
        mask_list.append(mask)

    mask = np.concatenate(mask_list, axis=0)
    init_weight = np.concatenate(init_weight_list, axis=0)
    init_bias = np.concatenate(bias_list, axis=0)
    return (
        mask.astype(np.float32),
        init_weight.astype(np.float32),
        init_bias.astype(np.float32),
    )


class build_layer_with_layer_parameter(nn.Module):
    def __init__(self, layer_parameters):
        super(build_layer_with_layer_parameter, self).__init__()

        os_mask, init_weight, init_bias = creak_layer_mask(layer_parameters)

        in_channels = os_mask.shape[1]
        out_channels = os_mask.shape[0]
        max_kernel_size = os_mask.shape[-1]

        self.weight_mask = nn.Parameter(torch.from_numpy(os_mask), requires_grad=False)

        self.padding = nn.ConstantPad1d(
            (int((max_kernel_size - 1) / 2), int(max_kernel_size / 2)), 0
        )

        self.conv1d = torch.nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=max_kernel_size,
        )
        self.conv1d.weight = nn.Parameter(
            torch.from_numpy(init_weight), requires_grad=True
        )
        self.conv1d.bias = nn.Parameter(torch.from_numpy(init_bias), requires_grad=True)

        self.bn = nn.BatchNorm1d(num_features=out_channels)

    def forward(self, X):
        self.conv1d.weight.data = self.conv1d.weight * self.weight_mask
        # self.conv1d.weight.data.mul_(self.weight_mask)
        result_1 = self.padding(X)
        result_2 = self.conv1d(result_1)
        result_3 = self.bn(result_2)
        result = F.relu(result_3)
        return result


class _OS_CNN(nn.Module):
    def __init__(
        self,
        n_class,  # number of classes to predict
        input_channel,
        receptive_field_shape,
        start_kernel_size=1,
        max_kernel_size=89,
        paramenter_number_of_layer_list=[8 * 128, 5 * 128 * 256 + 2 * 256 * 128],
        few_shot=False,
    ):
        super(_OS_CNN, self).__init__()
        self.few_shot = few_shot
        self.start_kernel_size = start_kernel_size
        self.max_kernel_size = max_kernel_size
        self.paramenter_number_of_layer_list = paramenter_number_of_layer_list
        self.input_channel = input_channel
        self.receptive_field_shape = receptive_field_shape

        layer_parameter_list = generate_layer_parameter_list(
            self.start_kernel_size,
            receptive_field_shape,
            self.paramenter_number_of_layer_list,
            in_channel=self.input_channel,
        )
        self.layer_list = []

        # building each model layer from a list of parameter
        for i in range(len(layer_parameter_list)):
            layer = build_layer_with_layer_parameter(layer_parameter_list[i])

            # print([layer.shape for layer in layer.parameters()])
            self.layer_list.append(layer)

        self.net = nn.Sequential(*self.layer_list)

        self.averagepool = nn.AdaptiveAvgPool1d(1)

        out_put_channel_numebr = 0
        for final_layer_parameters in layer_parameter_list[-1]:
            out_put_channel_numebr = out_put_channel_numebr + final_layer_parameters[1]

        self.hidden = nn.Linear(out_put_channel_numebr, n_class)

    def forward(self, X):
        X = self.net(X)

        X = self.averagepool(X)
        X = X.squeeze_(-1)

        if not self.few_shot:
            X = self.hidden(X)
        return X


class OSCNNCLassifier(SAILTorchClassifier):
    """
    Basic OS_CNN model

    Args:
        n_class: Number of classes to predict]
        input_channel: Number of input channels. varies depending on data type (univariate and multivariate)
        receptive_field_shape: Scales of receptive fields to avoid scale tuning
        start_kernel_size: Start kernel size. Must be a prime number
        max_kernel_size: Maximum kernel size. Must be a prime number
        paramenter_number_of_layer_list: OS block layer size
        device: cpu or gpu
        max_epochs: Epoch number
        batch_size: Size of the batch
        criterion: Loss function
        optimizer: Optimizer

    """

    def __init__(
        self,
        n_class,
        input_channel,
        receptive_field_shape,
        start_kernel_size=1,
        max_kernel_size=89,
        paramenter_number_of_layer_list=[8 * 128, 5 * 128 * 256 + 2 * 256 * 128],
        device="cpu",
        few_shot=False,
        max_epochs=20,
        **kwargs,
    ):
        super(OSCNNCLassifier, self).__init__(
            module=_OS_CNN,
            module__n_class=n_class,
            module__input_channel=input_channel,
            module__receptive_field_shape=receptive_field_shape,
            module__start_kernel_size=start_kernel_size,
            module__max_kernel_size=max_kernel_size,
            module__paramenter_number_of_layer_list=paramenter_number_of_layer_list,
            module__few_shot=few_shot,
            device=device,
            max_epochs=max_epochs,
            batch_size=16,
            criterion=nn.CrossEntropyLoss,
            optimizer=optim.Adam,
            **kwargs,
        )



# PyTorch implementation for LSTM FCN for Time Series Classification
# Original code in TensorFlow https://github.com/titu1994/LSTM-FCN
# Paper https://arxiv.org/abs/1709.05206
#
# By David Campos and Teodor Vernica

import torch
from torch import nn
from sail.models.torch import SAILTorchClassifier
from sail.models.torch.layers import (
    ConvBlock,
)  # The convolutional block using Conv1D with same padding.


class _LSTM_FCN(nn.Module):
    def __init__(
        self, in_channels: int, input_size: int, lstm_layers: int = 8, classes: int = 1
    ) -> None:
        """LSTM_FCN model.
        Args:
            in_channels (int): Number of input channels
            input_size (int): The input size
            lstm_layers (int): Number of hidden LSTM units
            classes (int): Number of classes
        """
        super().__init__()

        self.lstm = nn.LSTM(in_channels, 128, lstm_layers)
        self.h0 = torch.zeros(lstm_layers, input_size, 128)
        self.c0 = torch.zeros(lstm_layers, input_size, 128)
        self.drop = nn.Dropout(0.8)

        self.conv_layers = nn.Sequential(
            *[
                ConvBlock(in_channels, 128, 8, 1),
                ConvBlock(128, 256, 5, 1),
                ConvBlock(256, 128, 3, 1),
            ]
        )

        self.fc = nn.Linear(256, classes)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = torch.unsqueeze(x, 2)
        x_rnn, (hn, cn) = self.lstm(x, (self.h0, self.c0))
        x_rnn = self.drop(x_rnn)
        x_rnn = x_rnn[:, -1, :]

        x_cnn = self.conv_layers(x.transpose(2, 1))
        x_cnn = torch.mean(x_cnn, dim=-1)

        x_all = torch.cat((x_rnn, x_cnn), dim=1)
        x_out = self.softmax(self.fc(x_all))

        return x_out


class LSTMFCNClassifier(SAILTorchClassifier):
    def __init__(self, in_channels, input_size, lstm_layers, classes):
        """LSTM_FCN model.
        Args:
            in_channels (int): Number of input channels
            input_size (int): The input size
            lstm_layers (int): Number of hidden LSTM units
            classes (int): Number of classes
        """
        super(LSTMFCNClassifier, self).__init__(
            module=_LSTM_FCN,
            module__in_channels=in_channels,
            module__input_size=input_size,
            module__lstm_layers=lstm_layers,
            module__classes=classes,
            max_epochs=1,
        )



"""
Regressor for ensemble models
Based on: https://github.com/yzhao062/combo/blob/master/combo/models/base.py extended for
parallelizable incremental and batch learning algorithms.
"""
import numpy as np
import ray
from river import optim

from sail.utils.ray_utils import _model_fit

from .base import BaseAggregator

__all__ = ["DistAggregateRegressor"]


class DistAggregateRegressor(BaseAggregator):
    def __init__(
        self,
        estimators,
        fitted_estimators=[],
        loss=None,
        learning_rate=0.5,
        aggregator="simple",
    ):
        """
        :param estimators: Estimator objects with partial_fit defined
                    (incremental learning models).
        :param fitted_estimators: Batch learning estimators.
                Used only for scoring.
        :param aggregator: Type of aggregator. Options are "simple", "windsor" and "trim"
        """
        self.loss = optim.losses.Squared() if loss is None else loss
        self.learning_rate = learning_rate
        self.weights = [1.0] * len(estimators)
        self.aggregator = aggregator
        super().__init__(
            base_estimators=estimators, fitted_estimators=fitted_estimators
        )

    def _partial_fit(self, X, y=None, **kwargs):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :return:
        """
        obj_ids = []

        for i, estimator in enumerate(self.base_estimators):
            estimator.n_features_in_ = X.shape[1]
            X_id = ray.put(X)
            obj_ids.append(_model_fit.remote(estimator, X_id, y))
        self.base_estimators = ray.get(obj_ids)
        return self

    def partial_fit(self, X, y=None):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :return:
        """
        self._partial_fit(X, y)
        return self

    def predict(self, X):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :return: numpy array of shape (n_samples,).
             Predictions for input samples.
        """
        X_ensemble = []
        self.estimators = self.base_estimators
        self.estimators.extend(self.fitted_estimators)
        X_ensemble.append([model.predict(X) for model in self.estimators])
        X_ensemble = np.array(X_ensemble)
        # simple
        if self.aggregator == "simple":
            y_pred = np.nanmean(X_ensemble, axis=1)[0]
        # Windsorized
        if self.aggregator == "windsor":
            y_pred = (
                (
                    np.sum(X_ensemble, axis=1)
                    - np.amax(X_ensemble, axis=1)
                    - np.amin(X_ensemble, axis=1)
                    + np.sort(X_ensemble)[:, 1]
                    + np.sort(X_ensemble)[:, -2]
                )
                / (X_ensemble.shape[1])
            )[0]
        # Trimmed
        if self.aggregator == "trim":
            y_pred = (
                (
                    np.sum(X_ensemble, axis=1)
                    - np.amax(X_ensemble, axis=1)
                    - np.amin(X_ensemble, axis=1)
                )
                / (X_ensemble.shape[1] - 2)
            )[0]
        return y_pred



# Implementation of Time Series Classification from Scratch with Deep Neural Networks: A Strong Baseline (2016, arXiv) in PyTorch.
# The scripts are provided by https://github.com/aybchan/time-series-classification
import os

import numpy as np
import pandas as pd
import torch
from scipy.io.arff import loadarff
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm


class Data(Dataset):
    """
    Implements a PyTorch Dataset class
    """

    def __init__(self, dataset, testing):
        train_data = pd.DataFrame(loadarff(dataset + "_TRAIN.arff")[0])
        dtypes = {i: np.float32 for i in train_data.columns[:-1]}
        dtypes.update({train_data.columns[-1]: int})
        train_data = train_data.astype(dtypes)

        self.x = train_data.iloc[:, :-1].values
        self.y = train_data.iloc[:, -1].values

        self.mean = np.mean(self.x, axis=0)
        self.std = np.std(self.x, axis=0)
        if testing:
            test_data = pd.DataFrame(loadarff(dataset + "_TEST.arff")[0])
            dtypes = {i: np.float32 for i in test_data.columns[:-1]}
            dtypes.update({test_data.columns[-1]: int})
            test_data = test_data.astype(dtypes)

            self.x = test_data.iloc[:, :-1].values
            self.y = test_data.iloc[:, -1].values

        class_labels = np.unique(self.y)
        if (class_labels != np.arange(len(class_labels))).any():
            new_labels = {old_label: i for i, old_label in enumerate(class_labels)}
            self.y = [new_labels[old_label] for old_label in self.y]

        self.x = (self.x - self.mean) / self.std

    def __len__(self):
        return len(self.y)

    def __getitem__(self, idx):
        x = torch.Tensor([self.x[idx]]).float()
        y = torch.Tensor([self.y[idx]]).long()
        return x, y


def data_dictionary(dataset_path, datasets):
    """
    Create a dictionary of train/test DataLoaders for
    each of the datasets downloaded
    """
    dataset_dict = {}
    pbar = tqdm(datasets)
    print(pbar)
    for dataset in pbar:
        pbar.set_description("Processing {}".format(dataset))
        train_set, test_set = Data(os.path.join(dataset_path, dataset), False), Data(
            os.path.join(dataset_path, dataset), True
        )
        batch_size = min(16, len(train_set) // 10)

        dataset_dict[dataset] = {}
        dataset_dict[dataset]["train"] = DataLoader(train_set, batch_size=batch_size)
        dataset_dict[dataset]["test"] = DataLoader(test_set, batch_size=batch_size)

    return dataset_dict



import socket
from urllib.parse import urlparse


def validate_host_and_port(host, port, throw_exception=True):
    validation_success = True
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex((host, port))
        if result != 0:
            validation_success = False
            if throw_exception:
                raise Exception(f"Port: {port} is not open on the host: {host}")

    except socket.gaierror as e:
        validation_success = False
        if throw_exception:
            raise Exception(f"{host}: {str(e)}")
    finally:
        sock.close()

    return validation_success


def validate_address(address, throw_exception=True):
    endpoint = urlparse(address)
    return validate_host_and_port(endpoint.hostname, endpoint.port, throw_exception)
